/*
*  Name:      exceptions.hh
*  Author:    Rafael Jesus Alcantara Perez
*  Summary:   UESQLC exceptions
*  Date:      $Date: 2003-11-27 12:34:16 +0100 (jue, 27 nov 2003) $
*  Revision:  $Revision: 8538 $
*
*  Copyright (C) 1996-2002  Rafael Jesus Alcantara Perez <rafa@dedaloingenieros.com>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
*  MA 02111-1307, USA.
*/

#ifndef _EXCEPTIONS_UESQLC__
#define _EXCEPTIONS_UESQLC__

#include <ostream>
#include <mpcl/error_exception.hh>
#include <mpcl/exception.hh>
#include <mpcl/fatal_exception.hh>
#include <mpcl/warning_exception.hh>


/**
*  Exception: parent exception class for UESQLC.
*
*  Warning and errors should look like this:
*
*    source-file-name:lineno: message
*/
class TException: public virtual mpcl::TException 
{

  protected:

    //
    //  S E L E C T O R S
    //

    void write (std::basic_ostream<char>& rtTARGET_OSTREAM) const
    {
      if ( !fileName().empty() )
      {
        mpcl::TException::write (rtTARGET_OSTREAM);
      }
      else
      {
        rtTARGET_OSTREAM << generalDescription();
        if ( !specificDescription().empty() )
        {
          rtTARGET_OSTREAM << specificDescription();
        }
        rtTARGET_OSTREAM << std::endl;
      }
    }

};  // class TException


///  Exception: general warning.
class TWarningException                   :
  private virtual mpcl::TException        ,
  public          mpcl::TWarningException ,
  public          TException
{

  private:

    ///  Summary description.
    static const char*   _pkcSynopsis;


  public:

    //
    //  C O N S T R U C T O R S
    //

    /**
    *  Builds the instance with a generic and a (if used) specific description
    *  of the exception.
    *  @param rkyRATIONALE  Specific description.
    *  @param rkyFILE_NAME  File name where the exception has thrown.
    *  @param kiLINE_NUMBER Line number of the former file.
    */
    TWarningException ( const std::string& rkyRATIONALE  ,
                        const std::string& rkyFILE_NAME  ,
                        const int          kiLINE_NUMBER )                              :
      mpcl::TException          (_pkcSynopsis, rkyRATIONALE, rkyFILE_NAME, kiLINE_NUMBER) ,
      mpcl::TWarningException ()                                                        ,
      TException              ()                                                        {}

};  // class TWarningException


///  Exception: general error.
class TErrorException                   :
  private virtual mpcl::TException      ,
  public          mpcl::TErrorException ,
  public          TException
{

  private:

    ///  Summary description.
    static const char*   _pkcSynopsis;


  public:

    //
    //  C O N S T R U C T O R S
    //

    /**
    *  Builds the instance with a generic and a (if used) specific description
    *  of the exception.
    *  @param rkyRATIONALE  Specific description.
    *  @param rkyFILE_NAME  File name where the exception has thrown.
    *  @param kiLINE_NUMBER Line number of the former file.
    */
    TErrorException ( const std::string& rkyRATIONALE  ,
                      const std::string& rkyFILE_NAME  ,
                      const int          kiLINE_NUMBER )                              :
      mpcl::TException        (_pkcSynopsis, rkyRATIONALE, rkyFILE_NAME, kiLINE_NUMBER) ,
      mpcl::TErrorException ()                                                        ,
      TException            ()                                                        {}

};  // class TErrorException


/**
*  Exception: compiler runtime error.
*
*  Fatal errors should look like this:
*
*    program:source-file-name:lineno: message
*/
class TFatalException                   :
  private virtual mpcl::TException      ,
  public          mpcl::TFatalException ,
  public          TException
{

  private:

    ///  Summary description.
    static const char*   _pkcSynopsis;


  protected:

    //
    //  S E L E C T O R S
    //

    void write (std::basic_ostream<char>& rtTARGET_OSTREAM) const
    {
      //
      //  Fatal errors doesn't depend on files.
      //
      rtTARGET_OSTREAM << generalDescription();
      if ( !specificDescription().empty() )
      {
        rtTARGET_OSTREAM << specificDescription();
      }
      rtTARGET_OSTREAM << std::endl;
    }


  public:

    //
    //  C O N S T R U C T O R S
    //

    /**
    *  Builds the instance with a generic and a (if used) specific description
    *  of the exception.
    *  @param rkyRATIONALE  Specific description.
    *  @param rkyFILE_NAME  File name where the exception has thrown.
    *  @param kiLINE_NUMBER Line number of the former file.
    */
    TFatalException ( const std::string& rkyRATIONALE                  ,
                      const std::string& rkyFILE_NAME  = std::string() ,
                      const int          kiLINE_NUMBER = 0             )              :
      mpcl::TException        (_pkcSynopsis, rkyRATIONALE, rkyFILE_NAME, kiLINE_NUMBER) ,
      mpcl::TFatalException ()                                                        ,
      TException            ()                                                        {}

};  // class TFatalException


#endif  // not _EXCEPTIONS_UESQLC__
