/*
*  Name:      TAbstractConnectionManager.java
*  Author:    Rafael Jesus Alcantara Perez
*  Summary:   Class for handling connections in JAVA.
*  Date:      $Date: 2004-09-20 02:43:23 +0200 (lun, 20 sep 2004) $
*  Revision:  $Revision: 10141 $
*
*  Copyright (C) 2002-2003  Rafael Jesus Alcantara Perez <rafa@dedaloingenieros.com>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
*  MA 02111-1307, USA.
*/

///  Package for UESQL.
package org.uesqlc;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;


///  UESQL base manager class.
public abstract class TAbstractConnectionManager
{

  ///  Map that binds connection names to connections.
  protected Hashtable   tConnectionDataHashtable;

  ///  Current connection data (from \a tConnectionDataHashtable).
  protected TConnectionData   tCurrentConnectionData;

  ///  Input embedded variable string values queue.
  protected Vector   tInputEmbeddedVariableVector;

  ///  Output embedded variable string pointers vector.
  protected Vector   tOutputEmbeddedVariableVector;

  ///  String with the identifier of the current connection.
  protected String   yCurrentConnection;

  ///  Logger instance.
  public Logger   tLogger;


  //
  //  C O N S T R U C T O R S
  //

  /**
  *  Builds a new instance.  It uses a logger in the subsystem 'org.uesqlc' but
  *  it is deactivated by default.
  */
  public TAbstractConnectionManager()
  {

    tConnectionDataHashtable      = new Hashtable();
    tCurrentConnectionData        = null;
    tInputEmbeddedVariableVector  = new Vector();
    tOutputEmbeddedVariableVector = new Vector();
    yCurrentConnection            = null;
    tLogger                       = Logger.getLogger ("org.uesqlc");
    if ( tLogger.getLevel() == null )
    {
      //
      //  Deactivates logging if it has not been set externally.
      //
      tLogger.setLevel (Level.OFF);
    }
    initialize();

  }  // TAbstractConnectionManager()

  ///  Destroys the instance.
  public void terminate() throws SQLException
  {
    finish();
  }

  /**
  *  Binds in the  connection map,  a connection  named \a kyCONNECTION_NAME to
  *  \a tCONNECTION_DATA.
  *  @param kyCONNECTION_NAME Connection name.
  *  @param tCONNECTION_DATA  Connection data.
  */
  public void bindConnection (final String kyCONNECTION_NAME, TConnectionData tCONNECTION_DATA)
    throws SQLException
  {
    tConnectionDataHashtable.put (kyCONNECTION_NAME, tCONNECTION_DATA);
    tCONNECTION_DATA.getConnection().setAutoCommit (false);
  }

  /**
  *  Binds in the cursor map for the current connection, a cursor named
  *  \a kyCURSOR_NAME to \a tRESULT_SET.
  *  @param kyCURSOR_NAME Cursor name.
  *  @param tRESULT_SET   Result set instance.
  */
  public void bindCursor (final String kyCURSOR_NAME, ResultSet tRESULT_SET)
  {
    tCurrentConnectionData.bind (kyCURSOR_NAME, tRESULT_SET);
  }

  /**
  *  Commits current connection.
  *  @pre isConnected()
  */
  public void commitCurrent()
  {

    try
    {
      tCurrentConnectionData.getConnection().commit();
    }
    catch (final SQLException ktEXCEPTION)
    {
      //
      //  Condition:           invalid transaction termination
      //  Class subcondition:  no subclass
      //
      TState.SQLCODE    = -1;
      TState.SQLMESSAGE = ktEXCEPTION.getMessage();
      TState.SQLSTATE   = TState.invalid_transaction_termination._kyNO_SUBCLASS;
    }

  }  // commitCurrent()

  ///  Connects to current database, host and user login.
  public void connect (final String kyCONNECTION_NAME)
  {
    if ( tCurrentConnectionData != null )
    {
      connect ( kyCONNECTION_NAME                    ,
                tCurrentConnectionData.yHostName     ,
                tCurrentConnectionData.yCatalogName  ,
                tCurrentConnectionData.yUserName     ,
                tCurrentConnectionData.yUserPassword );
    }
    else
    {
      //
      //  Condition:           connection exception
      //  Class subcondition:  SQL-client unable to establish SQL-connection
      //
      TState.SQLCODE    = -1;
      TState.SQLSTATE   = TState.connection_exception._kySQLCLIENT_UNABLE_TO_ESTABLISH_SQLCONNECTION;
      TState.SQLMESSAGE = TState._getDescription();
    }

  }  // connect()

  /**
  *  Connects to current host and user login.
  *  @param kyCONNECTION_NAME Connection name.
  *  @param kyCATALOG_NAME    Catalog (database) name.
  */
  public void connect (final String kyCONNECTION_NAME, final String kyCATALOG_NAME)
  {

    if ( tCurrentConnectionData != null )
    {
      connect ( kyCONNECTION_NAME                    ,
                tCurrentConnectionData.yHostName     ,
                kyCATALOG_NAME                       ,
                tCurrentConnectionData.yUserName     ,
                tCurrentConnectionData.yUserPassword );
    }
    else
    {
      //
      //  Condition:           connection exception
      //  Class subcondition:  SQL-client unable to establish SQL-connection
      //
      TState.SQLCODE    = -1;
      TState.SQLSTATE   = TState.connection_exception._kySQLCLIENT_UNABLE_TO_ESTABLISH_SQLCONNECTION;
      TState.SQLMESSAGE = TState._getDescription();
    }

  }  // connect()

  /**
  *  Connects to a database.
  *  @param kyCONNECTION_NAME Connection name.
  *  @param kyHOST_NAME       Host name.
  *  @param kyCATALOG_NAME    Catalog (database) name.
  *  @param kyUSER_LOGIN      User login ('user/password').
  */
  public void connect ( final String kyCONNECTION_NAME ,
                        final String kyHOST_NAME       ,
                        final String kyCATALOG_NAME    ,
                        final String kyUSER_LOGIN      )
  {

    int      iBarIndex;
    String   yUserName;
    String   yUserPassword = "";

    iBarIndex = kyUSER_LOGIN.indexOf ('/');
    if ( iBarIndex == -1 )
    {
      yUserName = kyUSER_LOGIN.replaceAll ("'", "");
    }
    else
    {
      yUserName     = kyUSER_LOGIN.substring (0, iBarIndex).replaceAll ("'", "");
      yUserPassword = kyUSER_LOGIN.substring (iBarIndex + 1).replaceAll ("'", "");
    }
    connect ( kyCONNECTION_NAME ,
              kyHOST_NAME       ,
              kyCATALOG_NAME    ,
              yUserName         ,
              yUserPassword     );

  }  // connect()

  /**
  *  Connects to a database.
  *  @param kyCONNECTION_NAME Connection name.
  *  @param kyHOST_NAME       Host name.
  *  @param kyCATALOG_NAME    Catalog (database) name.
  *  @param kyUSER_NAME       User name.
  *  @param kyUSER_PASSWORD   User password.
  */
  public abstract void connect ( final String kyCONNECTION_NAME ,
                                 final String kyHOST_NAME       ,
                                 final String kyCATALOG_NAME    ,
                                 final String kyUSER_NAME       ,
                                 final String kyUSER_PASSWORD   );

  /**
  *  Undoes  all changes  made in the  current transaction and  in the  current
  *  connection.
  *  @pre isConnected()
  */
  public void rollbackCurrent()
  {

    try
    {
      tCurrentConnectionData.getConnection().rollback();
    }
    catch (final SQLException ktEXCEPTION)
    {
      //
      //  Condition:           invalid transaction termination
      //  Class subcondition:  no subclass
      //
      TState.SQLCODE    = -1;
      TState.SQLMESSAGE = ktEXCEPTION.getMessage();
      TState.SQLSTATE   = TState.invalid_transaction_termination._kyNO_SUBCLASS;
    }

  }  // rollbackCurrent()

  ///  Unbinds all connections.
  public void unbindAllConnections()
  {
    Enumeration   tElementEnumeration = tConnectionDataHashtable.keys();

    while ( tElementEnumeration.hasMoreElements() )
    {
      unbindConnection ((String) tElementEnumeration.nextElement());
    }
  }

  /**
  *  Unbinds connection named \a kyCONNECTION_NAME.
  *  @param kyCONNECTION_NAME Source connection name.
  *  @warning kyCONNECTION_NAME  mustn't  be  passed  as reference to avoid
  *           problems  about passing  \a yCurrentConnection  as  argument.
  */
  public void unbindConnection (final String kyCONNECTION_NAME)
  {

    TConnectionData   tConnectionData = (TConnectionData) tConnectionDataHashtable.get (kyCONNECTION_NAME);
    Connection        tConnection     = tConnectionData.getConnection();

    tConnectionDataHashtable.remove (kyCONNECTION_NAME);
    try
    {
      tConnection.commit();
      tConnection.close();
    }
    catch (final SQLException ktEXCEPTION)
    {
      //
      //  Condition:           connection exception
      //  Class subcondition:  connection failure
      //
      TState.SQLCODE    = -1;
      TState.SQLMESSAGE = ktEXCEPTION.getMessage();
      TState.SQLSTATE   = TState.connection_exception._kyCONNECTION_FAILURE;
    }
    if ( kyCONNECTION_NAME == yCurrentConnection )
    {
      yCurrentConnection = null;
    }

  }  // unbindConnection()

  ///  Unbinds current connection (if any).
  public void unbindCurrentConnection()
  {
    if ( isConnected() )
    {
      unbindConnection (yCurrentConnection);
    }
  }

  /**
  *  Unbinds cursor \a kyCURSOR_NAME from current connection.
  *  @param kyCURSOR_NAME Cursor name.
  */
  public void unbindCursor (final String kyCURSOR_NAME)
  {
    tCurrentConnectionData.unbind (kyCURSOR_NAME);
  }

  /**
  *  Clears  embedded  variable  containers.  It  must  be  called  at  the
  *  beginning of each new instruction.
  */
  public void clearQueues()
  {
    tInputEmbeddedVariableVector.clear();
    tOutputEmbeddedVariableVector.clear();
  }

  /**
  *  Executes the statement \a kySTATEMENT instantiating the input embedded
  *  variables (tagged with '$z') in the statement \a tSTATEMENT.
  *  @param tSTATEMENT  Statement where the string is executed.
  *  @param kySTATEMENT String with the source statement.
  */
  public ResultSet execute (final String kySTATEMENT)
  {

    String      yInstantiatedStatement;
    Statement   tStatement = null;
    ResultSet   tResultSet = null;

    resetStatus();
    try
    {
      yInstantiatedStatement = instantiate (kySTATEMENT);
      tStatement             = createStatement();
// <DEBUG>
      tLogger.info (yInstantiatedStatement);
// </DEBUG>
      if ( tStatement.execute (yInstantiatedStatement) )
      {
        tResultSet = tStatement.getResultSet();
      }
    }
    catch (final SQLException ktEXCEPTION)
    {
      //
      //  Condition:           transaction rollback
      //  Class subcondition:  statement completion unknown
      //
      TState.SQLCODE    = -1;
      TState.SQLMESSAGE = ktEXCEPTION.getMessage();
      TState.SQLSTATE   = TState.transaction_rollback._kySTATEMENT_COMPLETION_UNKNOWN;
      tResultSet        = null;
      try
      {
        rollbackCurrent();
        tStatement.close();
      }
      catch (final SQLException ktINNER_EXCEPTION) {}
    }
    if ( TState.SQLCODE == 0 )
    {
      if ( tResultSet == null )
      {
        // 
        //  Condition:  no data
        // 
        TState.SQLCODE    = 101;
        TState.SQLSTATE   = TState.no_data._kyNO_SUBCLASS;
        TState.SQLMESSAGE = TState._getDescription();
        try
        {
          tStatement.close();
        }
        catch (final SQLException ktEXCEPTION) {}
      }
    }
    return tResultSet;

  }  // execute()

  /**
  *  Executes the statement kySTATEMENT instantiating
  *  the input embedded variables (tagged with '$z'). The result
  *  of the execution is omitted.
  *  @param kySTATEMENT String with the source statement.
  */
  public void executeClean (final String kySTATEMENT)
  {

    String      yInstantiatedStatement;
    Statement   tStatement = null;

    resetStatus();
    try
    {
      yInstantiatedStatement = instantiate (kySTATEMENT);
      tStatement             = createStatement();
// <DEBUG>
      tLogger.info (yInstantiatedStatement);
// </DEBUG>
      tStatement.execute (yInstantiatedStatement);
      tStatement.close();
    }
    catch (final SQLException ktEXCEPTION)
    {
      try
      {
        tStatement.close();
      }
      catch (final SQLException ktINNER_EXCEPTION) {}
    }

  }  // executeClean()

  /**
  *  Executes  the  statement \a kySTATEMENT  instantiating the  input embedded
  *  variables (tagged with '$z') in the statement \a tSTATEMENT. The statement
  *  can not return more than one tuple.
  *  @param tSTATEMENT  Statement where the string is executed.
  *  @param kySTATEMENT String with the source statement.
  */
  public ResultSet executeSingleRow (final String kySTATEMENT)
  {

    String      yInstantiatedStatement;
    Statement   tStatement = null;
    ResultSet   tResultSet = null;

    resetStatus();
    try
    {
      yInstantiatedStatement = instantiate (kySTATEMENT);
      tStatement             = createStatement();
// <DEBUG>
      tLogger.info (yInstantiatedStatement);
// </DEBUG>
      tResultSet = tStatement.executeQuery (yInstantiatedStatement);
    }
    catch (final SQLException ktEXCEPTION)
    {
      //
      //  Condition:           transaction rollback
      //  Class subcondition:  statement completion unknown
      //
      TState.SQLCODE    = -1;
      TState.SQLMESSAGE = ktEXCEPTION.getMessage();
      TState.SQLSTATE   = TState.transaction_rollback._kySTATEMENT_COMPLETION_UNKNOWN;
      tResultSet        = null;
      try
      {
        rollbackCurrent();
        tStatement.close();
      }
      catch (final SQLException ktINNER_EXCEPTION) {}
    }
    if ( TState.SQLCODE == 0 )
    {
      try
      {
        if ( !tResultSet.next() )
        {
          // 
          //  Condition:  no data
          // 
          TState.SQLCODE    = 101;
          TState.SQLSTATE   = TState.no_data._kyNO_SUBCLASS;
          TState.SQLMESSAGE = TState._getDescription();
          tResultSet        = null;
        }
        else
        {
          if ( !tResultSet.next() )
          {
            //
            //  There is only one row, so everything is all right.
            //
            tResultSet.beforeFirst();
          }
          else
          {
            // 
            //  Condition:  cardinality violation
            // 
            TState.SQLCODE    = -1;
            TState.SQLSTATE   = TState.cardinality_violation._kyNO_SUBCLASS;
            TState.SQLMESSAGE = TState._getDescription();
            tResultSet        = null;
            try
            {
              rollbackCurrent();
              tStatement.close();
            }
            catch (final SQLException ktINNER_EXCEPTION) {}
          }
        }
      }
      catch (final SQLException ktEXCEPTION)
      {
        //
        //  Condition:           invalid cursor state
        //  Class subcondition:  no subclass
        //
        TState.SQLCODE    = -1;
        TState.SQLMESSAGE = ktEXCEPTION.getMessage();
        TState.SQLSTATE   = TState.invalid_cursor_state._kyNO_SUBCLASS;
        tResultSet        = null;
        try
        {
          rollbackCurrent();
          tStatement.close();
        }
        catch (final SQLException ktINNER_EXCEPTION) {}
      }
    }
    return tResultSet;

  }  // executeSingleRow()

  ///  Finishes all managed issues.
  public void finish() throws SQLException
  {
    unbindAllConnections();
    clearQueues();
    tConnectionDataHashtable.clear();
    yCurrentConnection = null;
  }

  /**
  *  Gets a pointer to the SQL data type of the output embedded variable at
  *  position kiINDEX in the  output embedded variable vector.  It  must be
  *  called for writing results on output embedded variables.
  *  @param kiINDEX Position of the variable.
  *  @return Pointer to the output embedded variable.
  */
  public IType getFromOutputQueue (final int kiINDEX)
  {
    return (IType) tOutputEmbeddedVariableVector.get (kiINDEX);
  }

  ///  Initializes the default connection.
  public void initialize()
  {
    connect ("DEFAULT");
    if ( !isSqlerrorCondition() )
    {
      setCurrentConnection ("DEFAULT");
    }
  }

  /**
  *  Returns a string that is result of instanting the input
  *  embedded variables (if any).
  *  @param kySTATEMENT Source statement.
  *  @return The instantiated string.
  */
  public String instantiate (final String kySTATEMENT)
  {

    int            iPatternIndex;
    IType          tType;
    StringBuffer   yStatementInstance = new StringBuffer (kySTATEMENT);

    while ( !tInputEmbeddedVariableVector.isEmpty() )
    {
      iPatternIndex = yStatementInstance.indexOf ("$z");
      if ( -1 == iPatternIndex )
      {
        break;
      }
      else
      {
        tType = (IType) tInputEmbeddedVariableVector.firstElement();
        yStatementInstance.replace (iPatternIndex, iPatternIndex + 2, tType.get());
        tInputEmbeddedVariableVector.remove (0);
      }
    }
    return yStatementInstance.toString();

  }  // instantiate()

  /**
  *  Pops from the output embedded variable queue the last entered variable and
  *  returns it.
  *  @return The last entered value.
  */
  public IType popFromOutputQueue()
  {
    return (IType) tOutputEmbeddedVariableVector.remove (tOutputEmbeddedVariableVector.size() - 1);
  }

  /**
  *  Pushes on the input embedded variable queue the variable
  *  value \a ktVALUE.
  *  @param ktVALUE Input embedded variable value.
  */
  public void pushOnInputQueue (final IType ktVALUE)
  {
    tInputEmbeddedVariableVector.add (ktVALUE);
  }

  /**
  *  Pushes on the output embedded variable vector the variable
  *  reference \a tVARIABLE.
  *  @param tVARIABLE Output embedded variable reference.
  */
  public void pushOnOutputQueue (IType tVARIABLE)
  {
    tOutputEmbeddedVariableVector.add (tVARIABLE);
  }

  ///  Resets SQL status variables.
  public void resetStatus()
  {
    //
    //  Condition:  successful completion
    //
    TState.SQLCODE    = 0;
    TState.SQLSTATE   = TState.successful_completion._kyNO_SUBCLASS;
    TState.SQLMESSAGE = TState._getDescription();
  }

  /**
  *  Sets current connection to connection named kyCONNECTION_NAME.
  *  @param kyCONNECTION_NAME Source connection name.
  */
  public void setCurrentConnection (final String kyCONNECTION_NAME)
  {

    if ( tConnectionDataHashtable.containsKey (kyCONNECTION_NAME) )
    {
      yCurrentConnection     = kyCONNECTION_NAME;
      tCurrentConnectionData = (TConnectionData) tConnectionDataHashtable.get (kyCONNECTION_NAME);
// <DEBUG>
      tLogger.info ("SET CONNECTION '" + kyCONNECTION_NAME + "'");
// </DEBUG>
      resetStatus();
    }
    else
    {
      //
      //  Condition:           connection exception
      //  Class subcondition:  connection does not exist
      //
      TState.SQLCODE    = -1;
      TState.SQLSTATE   = TState.connection_exception._kyCONNECTION_DOES_NOT_EXIST;
      TState.SQLMESSAGE = TState._getDescription();
    }
      
  }  // setCurrentConnection()


  //
  //  S E L E C T O R S
  //

  /**
  *  Creates a new statement for the current connection.
  *  @return The newly created statement.
  */
  protected Statement createStatement() throws SQLException
  {
    return tCurrentConnectionData.getConnection().createStatement();
  }

  /**
  *  Returns the connection name for the current connection.
  *  @return The connection name for the current connection.
  */
  public final String currentConnectionName()
  {
    return yCurrentConnection;
  }

  /**
  *  Gets cursor named \a kyCURSOR_NAME.
  *  @param kyCURSOR_NAME Cursor name.
  *  @return Result set.
  */
  public ResultSet getCursor (final String kyCURSOR_NAME)
  {
    return tCurrentConnectionData.getCursor (kyCURSOR_NAME);
  }

  /**
  *  Returns true if there is a default connection.
  *  @return True if there is a default connection.
  */
  public boolean isConnected()
  {
    return ( yCurrentConnection != null );
  }

  /**
  *  Returns true if there is an active connection.
  *  @param kyCONNECTION_NAME Source connection name to check.
  *  @return True if there is an active connection.
  */
  public boolean isConnected (final String kyCONNECTION_NAME)
  {
    return tConnectionDataHashtable.containsKey (kyCONNECTION_NAME);
  }

  /**
  *  Checks if a NOT FOUND condition has happened.
  *  @return True if it has happened.
  */
  public boolean isNotFoundCondition()
  {
    return ( TState.SQLSTATE == TState.no_data._kyNO_SUBCLASS );
  }

  /**
  *  Checks if a SQLERROR condition has happened.
  *  @return True if it has happened.
  */
  public boolean isSqlerrorCondition()
  {
    return ( TState.SQLCODE < 0 );
  }

}  // class TAbstractConnectionManager
